# -*- coding: utf-8 -*-
"""
Philipp-Maximilian Jacob, Peyman Yamin, Charles Perez-Storey, Matt Hopgood, Alexei Lapkin
Institution: University of Cambridge, Department of Chemical Engineering and Biotechnology

All rights reserved

Description:

This script calculates the E factor calculation.

It is designed for a linear synthesis plan with variable 
 - number of steps
 - number of reagents
 - number of byproducts
 - stoichiometry.

Assumptions:
 - The limiting reagent is the desired product from a previous step (i.e. main organic reagent)
 - The limiting reagent is listed as first reagent
 - All desired product is ideally separated from one step to the next

Steps undertaken:
    1) Prepare script for algorithm
    2) Set Directory (HARD-CODED)
    3) Import Data from CSV (CSV FORMAT IS HARD-CODED)
    4) Extract data to various arrays
    5) Calculate kernel parameters
    6) Import Routes from CSV (CSV FORMAT IS HARD-CODED)
    7) Finish E Factor Calculations
    8) Output Individual and Summary Files (OUTPUT FORMAT IS HARD CODED)

File inputs:
 - "EfactorData.dat" - which contains all the data on the reactions following 
   the template provided. The first reactant listed is assumed to be the main
   and thus limiting reactant.
 - "NetworkRoutes.dat" - a file containing the list of possible routes through
   the network
File outputs of this script:

EFactorSummary.csv                      
 - Summary of E Factors for all routes calculated. 
 - Contains breakdown of contribution from kernel, excess, auxiliaries.

Route "FirstRxnID"-...-"FinalRxnID".csv
 - Filename constructed of the sequence of reactions within the route. 
 - Contains parameters broken down by reaction in that route.
 - Individual file per route calculated.
"""

#%% Step 0: Prepare Script for Algorithm
"""
-------------------------------------------------------------------------------
Step 0: Prepare script for algorithm
-------------------------------------------------------------------------------
"""
#This cell imports relevant packages for use in subsequent cells
import csv
import re
import os
import numpy as np
import uncertainties as unct
from operator import itemgetter

def unctarray(a,b):
    i=0
    temp = np.array([])
    while i<a*b:
        temp=np.append(temp,np.array([unct.ufloat(0,0)]))
        i+=1
    array = np.reshape(temp,(a,b))
    return(array)

#%% Set Directory

#Ensures Working Directory is in correct place to find data file
current_path = os.path.abspath(__file__)
current_path = os.path.split(current_path)
current_path = current_path[0]
os.chdir(current_path)
# =======================================================================================================
#os.chdir(r'C:\Users\Charles\Desktop\Research Project Paper Files')     # DECLARE WORKING DIRECTORY HERE
# =======================================================================================================

#
#
#
#
#
#
#
#
#
#%% Import CSV Data: 1 CSV
"""
-------------------------------------------------------------------------------
Step 1: Import CSV Data

Reaction data should be presented as a CSV into the working directory in the 
style as outlined in the template file.
-------------------------------------------------------------------------------
"""

# Extract Data from CSV File Into Nested Lists
MainDataList=[]
with open ("EfactorData.dat", "r") as datfile:
    reader = csv.reader(datfile, delimiter=',')
    for row in reader:
        temp = []
        for entry in row:
            entry = entry.strip()
#            if '+-' in element:
#                element = unct.ufloat_fromstr(element)
            temp.append(entry)
        MainDataList.append(temp)
#        MainDataList.append([element.strip() for element in row])

for rxnlist in MainDataList:
    for k, val in enumerate(rxnlist):
        rxnlist[k] = re.split('#',val)

TotalReactionNumber = len(MainDataList) - 1
MaxVertIndex = len(MainDataList)
TotalParamNumber = len(MainDataList[0])

#        
#%%
"""
-------------------------------------------------------------------------------
Step 2: Extract Data From Main Extraction to Numpy Arrays
-------------------------------------------------------------------------------
"""
# Parameters
Params = MainDataList[0]
Paramarray = np.array(Params).transpose()

""" STEP 1: Initialize Arrays """
    
    # RXN ID
RXN_IDS = np.zeros((1,TotalReactionNumber))
RXN_IDS_LIST = []

"""____Products____ """
    
# Count up number of number of reagents, and string sizes for array initialization
P_Name_Size = np.zeros((1,TotalReactionNumber))
B_Number = np.zeros((1,TotalReactionNumber))
B_Name_Size = np.zeros((1,TotalReactionNumber))

for j in range(0,TotalReactionNumber):
    P_Name_Size[0,j] = len(MainDataList[j + 1][1][0])
    B_Number[0,j] = len(MainDataList[j+1][4])
    for k in range(0,int(B_Number[0,j])):
        BStringLength = len(MainDataList[j+1][4][k])
        if BStringLength > B_Name_Size[0,j]:
            B_Name_Size[0,j] = BStringLength
    
MaxPNameSize = int(np.amax(P_Name_Size))
MaxBNumber = int(np.amax(B_Number))
MaxBNameSize = int(np.amax(B_Name_Size))

# Initialize Arrays
P_j_Name = np.zeros((1,TotalReactionNumber),dtype = '|S'+str(MaxPNameSize))
P_j_MW = np.zeros((1,TotalReactionNumber))
GAMMA_j = np.zeros((1,TotalReactionNumber))
B_ij_Name = np.zeros((MaxBNumber,TotalReactionNumber),dtype = '|S'+str(MaxBNameSize))
B_ij_MW = np.zeros((MaxBNumber,TotalReactionNumber))
BETA_ij = np.zeros((MaxBNumber,TotalReactionNumber))

"""____Reagents____"""

# Count up number of number of reagents, and string sizes for array initialization
R_Number = np.zeros((1,TotalReactionNumber))
R_Name_Size = np.zeros((1,TotalReactionNumber))
for j in range(0,TotalReactionNumber):              #REAGENTS - SIMILAR METHOD AS BYPRODUCTS BECAUSE WE NEED VARIABLE SIZE ARRAYS, 
    R_Number[0,j] = len(MainDataList[j + 1][7])     # REPLICATE COUNTING STRUCTURE YOU HAVE PREVIOUSLY
    for k in range(0,int(R_Number[0,j])):
        RStringLength = len(MainDataList[j + 1][7][k]) # POINT FOR IMPROVEMENT: ERROR CHECKING THE DATA TO MAKE SURE THERE ARE THE SAME 
        if RStringLength > R_Name_Size[0,j]:           # NUMBER OF DATA POINTS FOR EACH REAGENT, FLAGGING IF THIS IS NOT THE CASE,
            R_Name_Size[0,j] = RStringLength           # DETERMINATION OF ARRAY SIZES BY RECOUNTING EACH INDIVIDUAL DATA SUBSET

MaxRNumber = int(np.amax(R_Number))
MaxRNameSize = int(np.amax(R_Name_Size))

# Initialize Arrays
R_ij_Name = np.zeros((MaxRNumber,TotalReactionNumber),dtype = '|S'+str(MaxRNameSize))
R_ij_MW = unctarray(MaxRNumber,TotalReactionNumber)
ALPHA_j = unctarray(MaxRNumber,TotalReactionNumber)
THETA_j = unctarray(MaxRNumber,TotalReactionNumber)

"""____Auxiliaries____"""
EPSILON_j = unctarray(1,TotalReactionNumber)
c_j = unctarray(1,TotalReactionNumber)
s_j = unctarray(1,TotalReactionNumber)
w_j=unctarray(1,TotalReactionNumber)

""" STEP 2: Construct arrays and fill cells """
for j in range(0,TotalReactionNumber):
    RXN_IDS[0,j] = MainDataList[j + 1][0][0]            # RXN IDs
    RXN_IDS_LIST.append(MainDataList[j + 1][0][0])
    P_j_Name[0,j] = (MainDataList[j + 1][1][0])         # Desired Product Name
    P_j_MW[0,j] = MainDataList[j + 1][2][0]             # Desired Product MW
    GAMMA_j[0,j] = MainDataList[j + 1][3][0]            # Desired Product Stoichiometry
    for k in range(0,int(B_Number[0,j])):               
        B_ij_Name[k,j] = MainDataList[j + 1][4][k]      # Byproduct Names
        B_ij_MW[k,j] = MainDataList[j + 1][5][k]        # Byproduct MWs
        BETA_ij[k,j] = MainDataList[j + 1][6][k]        # Byproduct Stoichiometries
    for k in range(0,int(R_Number[0,j])):
        R_ij_Name[k,j] = MainDataList[j + 1][7][k]      # Reagent Names
        R_ij_MW[k,j] = unct.ufloat_fromstr(MainDataList[j + 1][8][k])        # Reagent MWs
        ALPHA_j[k,j] = unct.ufloat_fromstr(MainDataList[j + 1][9][k])        # Reagent Stoichiometries
        THETA_j[k,j] = unct.ufloat_fromstr(MainDataList[j + 1][10][k])      # Reagent Moles
    EPSILON_j[0,j] = unct.ufloat_fromstr(MainDataList[j + 1][11][0])         # Yield
    c_j[0,j] = unct.ufloat_fromstr(MainDataList[j + 1][12][0])               # Catalyst Mass
    s_j[0,j] = unct.ufloat_fromstr(MainDataList[j + 1][13][0])               # Solvent Mass
    w_j[0,j] = unct.ufloat_fromstr(MainDataList[j + 1][14][0])               # Workup Mass
    
#
#
#
#
#
#
#
#
#
#
#
#%% Run through Calculations
"""
-------------------------------------------------------------------------------
Step 3: Calculate parameters for individual steps (kernel level)
-------------------------------------------------------------------------------
"""
## Calculations for parameters that are independent of route being chosen

# AEj (Determine AE for each reaction step)
#AEj = np.zeros((1,TotalReactionNumber))
AEj = np.array([])
i=0
temp = np.array([])
while i<TotalReactionNumber*MaxRNumber:
    temp=np.append(temp,np.array([unct.ufloat(0,0)]))
    i+=1
AEj = np.reshape(temp,(MaxRNumber,TotalReactionNumber))
#RStoichSUM = np.zeros((MaxRNumber,TotalReactionNumber))
RStoichSUM = np.array([])
i=0
temp = np.array([])
while i<TotalReactionNumber*MaxRNumber:
    temp=np.append(temp,np.array([unct.ufloat(0,0)]))
    i+=1
RStoichSUM = np.reshape(temp,(MaxRNumber,TotalReactionNumber))
for j in range(0,TotalReactionNumber,1): 
    for k in range(0,int(R_Number[0,j])):
        RStoichSUM[k,j] = (ALPHA_j[k,j] * R_ij_MW[k,j]) 
    AEj[0,j] = P_j_MW[0,j] * GAMMA_j[0,j] / np.sum(RStoichSUM[:,j])       

# Determine limiting reagent and stoichiometry for each step
Xjstar = unctarray(1,TotalReactionNumber)              # Moles of Limiting Reagent
Ajstar = unctarray(1,TotalReactionNumber)             # Stoichiometry of Limiting Reagent
for j in range(0,TotalReactionNumber,1):   
    Xjstar[0,j] = THETA_j[0,j]
    Ajstar[0,j] = ALPHA_j[0,j]

# Calculate SFj (Stoichiometric factor)
SFj = unctarray(1,TotalReactionNumber)
RStoichMass = unctarray(MaxRNumber,TotalReactionNumber)
RActualMass = unctarray(MaxRNumber,TotalReactionNumber)
RExcessMass = unctarray(MaxRNumber,TotalReactionNumber)
for j in range(0,TotalReactionNumber):    
    for k in range(0,int(R_Number[0,j])):    
        RStoichMass[k,j] = Xjstar[0,j] * ALPHA_j[k,j] * R_ij_MW[k,j] / Ajstar[0,j]
        RActualMass[k,j] = THETA_j[k,j] * R_ij_MW[k,j]
        RExcessMass[k,j] = RActualMass[k,j] - RStoichMass[k,j]
    SFj[0,j] = 1 + np.sum(RExcessMass[:,j]) / np.sum(RStoichMass[:,j])
#
#
#
#
#
#   
#  
#   
#   
#   
#%% Choose reaction set
"""
-------------------------------------------------------------------------------
Step 4: Declare reaction set to be analysed
-------------------------------------------------------------------------------
"""

""" Import reaction set """
# Imports Reaction Routes from NetworkRoutes.dat file. 
# N.b. Reaction IDs must match between NetworkRoutes.dat and EFactorData4.dat else an error will result
# There is NO AUTO-CONSISTENCY CHECKING between import of reaction routes and the EFactor4.dat file.
ReactionIDset=[]
with open ("NetworkRoutes.dat", "r") as datfile:
    reader = csv.reader(datfile, delimiter='\t')
    for row in reader:
        ReactionIDset.append([element.strip() for element in row])

# ================================================================================================================================
# ReactionIDset = [[1.1,2.1,3.2],[1.2,2.1,3.2],[1.3,2.1,3.2]]     # HARD CODE ALTERNATIVE: DECLARE ROUTES SET FOR CALCULATION HERE
# ================================================================================================================================

# Create Index List from ReactionIDset
ReactionINDEXset = []
i = 0
IDS_transposed = np.transpose(RXN_IDS)
for routeID in ReactionIDset:
    ReactionINDEXset.append([])    
    for rxnID in routeID:
        try:
            rxnIDindex = IDS_transposed.tolist().index([int(rxnID)])
        except ValueError:
            rxnIDindex = IDS_transposed.tolist().index([float(rxnID)])            
        ReactionINDEXset[i].append(rxnIDindex)
    i = i + 1  
    
# Determine Length of Final Arrays Required
RouteNumber = len(ReactionINDEXset)
Routes = []

#%% Calculation of green metrics
"""
-------------------------------------------------------------------------------
Step 5: Calculate parameters dependent on reaction set
-------------------------------------------------------------------------------
"""

# Prepare Final Arrays
Etotal = unctarray(1,RouteNumber)
Ekernel = unctarray(1,RouteNumber)
Eexcess = unctarray(1,RouteNumber)
Eaux = unctarray(1,RouteNumber)

""" Cycle through Reaction Set """
for route in ReactionINDEXset:
    routeindex = ReactionINDEXset.index(route)
    SyntreeLength = len(ReactionINDEXset[ReactionINDEXset.index(route)])
    Finalindex = SyntreeLength - 1
    
    """ Organise Reaction Parameters """
    #Initialise shorter parameter sets
    RxnIDS = np.zeros((1,SyntreeLength))
    RxnP_j_MW = np.zeros((1,SyntreeLength))
    RxnAEj = unctarray(1,SyntreeLength)
    RxnSFj = unctarray(1,SyntreeLength)
    RxnC_j = unctarray(1,SyntreeLength)
    RxnS_j = unctarray(1,SyntreeLength)
    RxnW_j = unctarray(1,SyntreeLength)
    RxnXjstar = unctarray(1,SyntreeLength)
    RxnAjstar = unctarray(1,SyntreeLength)
    RxnEPSILON_j = unctarray(1,SyntreeLength)
    RxnGAMMA_j = np.zeros((1,SyntreeLength))
    RxnALPHA_j = unctarray(1,SyntreeLength)        
   
   # Fill new parameter sets
    i = 0
    for rxn in route:
        RxnIDS[0,i] = RXN_IDS[0,rxn]
        RxnP_j_MW[0,i] = P_j_MW[0,rxn]
        RxnAEj[0,i] = AEj[0,rxn]
        RxnSFj[0,i] = SFj[0,rxn]
        RxnC_j[0,i] = c_j[0,rxn]
        RxnS_j[0,i] = s_j[0,rxn]
        RxnW_j[0,i] = w_j[0,rxn]
        RxnXjstar[0,i] = Xjstar[0,rxn]
        RxnAjstar[0,i] = Ajstar[0,rxn]            
        RxnEPSILON_j[0,i] = EPSILON_j[0,rxn]
        RxnGAMMA_j[0,i] = GAMMA_j[0,rxn]
        RxnALPHA_j[0,i] = ALPHA_j[0,rxn]         
        
        i = i + 1

    """ MULTISUMS """
    # Multiplicative sum of yields
    MultisumEPSILON = unctarray(1,SyntreeLength)            
    templength = 1   
    for i in range(Finalindex, -1,-1):
        temparray = unctarray(templength,1)
        counter = Finalindex
        for k in range(0,templength,1):
            temparray[k,0] = RxnEPSILON_j[0,counter]
            counter = counter - 1        
        MultisumEPSILON[0,i] = temparray.prod() 
        templength = templength + 1   
  
    # Multiplicative sum of Limiting Reagent Stoichiometry (Ajstar)   
    MultisumAjstar = unctarray(1,SyntreeLength)         
    templength = 1   
    for i in range(Finalindex, -1,-1):
        temparray = unctarray(templength,1)
        counter = Finalindex
        for k in range(0,templength,1):
            temparray[k,0] = RxnAjstar[0,counter]
            counter = counter - 1        
        MultisumAjstar[0,i] = temparray.prod() 
        templength = templength + 1 
        
    # Multiplicative sum of main organic reagent stoichiometry (ALPHAs)
    MultisumALPHA = unctarray(1,SyntreeLength)            
    templength = 1   
    for i in range(Finalindex, -1,-1):
        temparray = unctarray(templength,1)
        counter = Finalindex
        for k in range(0,templength,1):
            temparray[k,0] = RxnALPHA_j[0,counter]
            counter = counter - 1        
        MultisumALPHA[0,i] = temparray.prod() 
        templength = templength + 1 
        
    """ E Factor Calc """
    # ETotal Sum Elements
    Etotalsum = unctarray(1,SyntreeLength)
    
    Ekernelsum = unctarray(1,SyntreeLength)
    Eexcesssum = unctarray(1,SyntreeLength)
    Eauxsum = unctarray(1,SyntreeLength)
    
    for j in range(0,SyntreeLength,1):
        Etotalsum[0,j] = ( (1)/(MultisumEPSILON[0,j]) ) * ( (RxnP_j_MW[0,j]/RxnAEj[0,j])* (RxnGAMMA_j[0,j]/RxnAjstar[0,j]) * (                   
                        RxnSFj[0,j] - (RxnAjstar[0,j]/RxnGAMMA_j[0,j]) * RxnEPSILON_j[0,j]*RxnAEj[0,j])
                        + (RxnC_j[0,j]+RxnS_j[0,j]+RxnW_j[0,j])/RxnXjstar[0,j] 
                        )
        
        Ekernelsum[0,j] = ( 1/MultisumEPSILON[0,j] ) * ( (RxnP_j_MW[0,j]/RxnAEj[0,j])*(RxnGAMMA_j[0,j]/RxnAjstar[0,j]) * (        
                         1 - (RxnAjstar[0,j]/RxnGAMMA_j[0,j])*RxnEPSILON_j[0,j]*RxnAEj[0,j] ) )
                        
        Eexcesssum[0,j] = ( 1/MultisumEPSILON[0,j] ) * ( (RxnP_j_MW[0,j]/RxnAEj[0,j])*(RxnGAMMA_j[0,j]/RxnAjstar[0,j]) * (
                         RxnSFj[0,j] - 1 ) )
        
        Eauxsum[0,j] = ( 1/MultisumEPSILON[0,j] ) * ( (RxnC_j[0,j]+RxnS_j[0,j]+RxnW_j[0,j])/RxnXjstar[0,j] )
                
                
    # Final Summation
                
    Etotal[0,routeindex] = (1 / RxnP_j_MW[0,Finalindex]) * np.sum(Etotalsum)
       
    Ekernel[0,routeindex] = (1 / RxnP_j_MW[0,Finalindex]) * np.sum(Ekernelsum)
    Eexcess[0,routeindex] = (1 / RxnP_j_MW[0,Finalindex]) * np.sum(Eexcesssum)
    Eaux[0,routeindex] = (1 / RxnP_j_MW[0,Finalindex]) * np.sum(Eauxsum)
    
    
    """ Output Parameters to File """
    # Choose File Name
    FileName = 'Route '
    RouteName = ''    
    for i in range(len(route)):
        if i == len(route) - 1:
            FileName += str(RXN_IDS_LIST[route[i]])
            RouteName += str(RXN_IDS_LIST[route[i]])            
        elif i < len(route) - 1:
            FileName += str(RXN_IDS_LIST[route[i]])+'-'
            RouteName += str(RXN_IDS_LIST[route[i]])+'-'
    
    Routes.append(RouteName)
    
    # Write Parameters to File            
    VarWriteList = [RxnIDS,RxnXjstar,RxnAjstar,MultisumALPHA,MultisumEPSILON,RxnSFj,RxnAEj]
    VarNamesWriteList = ['Rxn ID','Moles Limiting Reagent (Xjstar)','Stoichiometry Limiting Reagent (Ajstar)','MultisumALPHA','MultisumEPSILON','SFj','AEj']
    rowWriteList = []    
    for var_index in range(len(VarWriteList)):
        temprow = []
        temprow.append(VarNamesWriteList[var_index])
        for i in range(len(route)):
            temprow.append(VarWriteList[var_index][0,i])
        rowWriteList.append(temprow)
       
    with open(FileName+'.csv','wb') as csvfile:
        
        EfactorWriter = csv.writer(csvfile, dialect='excel', delimiter=',')
        for row in rowWriteList:    
            EfactorWriter.writerow(row)

#%% Output
"""
-------------------------------------------------------------------------------
Step 6: Output Parameters to Summary File
-------------------------------------------------------------------------------
"""

""" Write Main E Factors to ONE FILE """

#sort output
sort_list = []
for i in range(0,RouteNumber):
    temp= []
    temp=[Etotal[0,i],Routes[i]]
    sort_list.append(temp)

sort_list.sort(key=lambda sublist: sublist[0])
# Prepare row lists for writing
Route_row = ['Route']
Etotal_row = ['Etotal']
for path in sort_list:
    
    Route_row.append(path[0])
    Etotal_row.append(path[1])
# Choose Output Parameters
output_params = [Route_row,Etotal_row]      #,Ekernel_row,Eexcess_row,Eaux_row]

# Write parameters to file
with open('EFactorSummary.csv','wb') as summaryfile:  
    writer = csv.writer(summaryfile, dialect='excel', delimiter=',')
    for var in output_params:
        writer.writerow(var)
    
"""
********** END OF SCRIPT ********** END OF SCRIPT ********** END OF SCRIPT ********** END OF SCRIPT **********
"""